@@ -40,8 +40,8 @@
 defines
 *****************************************************************/
 
-#define SAB_VERSION "0.44"
-#define SAB_DATE "September 13, 2000"
+#define SAB_VERSION "0.51"
+#define SAB_DATE "February 19, 2001"
 #define BARE_SIGN '@'
 
 #define MAX_BARE        3
@@ -574,7 +574,3 @@ int main(int argc, char *argv[])
     return 0;
 }
 
-
-
-
-
