@@ -32,10 +32,10 @@
 
 #include "datastr.h"
 #include "verts.h"
-#include <ctype.h>
+#include "utf8.h"
 // #include <malloc.h> - not needed and deprecated anyway
 
-const Str theEmptyString("");
+/* const Str* theEmptyString = new Str(""); */
 
 //     implementations for List are in datastr.h (List is a template)
 
@@ -62,7 +62,7 @@ inline void DynBlock::remove()
         {
             d_was = d;
             d = d_was -> next;
-            delete d_was -> data;
+            delete [] d_was -> data;
             delete d_was;
         }
     };
@@ -280,7 +280,7 @@ Str& Str::operator= (double num)
 {
     remove_();
     char buf[20];
-    sprintf(buf,"%g",num);
+    sprintf(buf,"%.13g",num);
     operator=(buf);
     return *this;
 }
@@ -307,12 +307,12 @@ Bool Str::isEmpty() const
     return !length(); // this is virtual
 }
 
-inline int Str::length() const
+int Str::length() const
 {
     return byteLength_;
 }
 
-inline Str::operator char*() const
+Str::operator char*() const
 {
     if (!text_)
         const_cast<Str*>(this) -> empty();
@@ -328,7 +328,7 @@ void Str::empty()
 
 int Str::compare(const char* otherChars) const
 {
-    return (strcmp(operator char*(), otherChars) < 0);
+    return (strcmp(operator char*(), otherChars));
 }
 
 int Str::compare(const Str& other) const
@@ -346,6 +346,11 @@ Bool Str::operator== (const char* otherChars) const
     return (Bool) !strcmp(operator char*(), otherChars);
 }
 
+Bool Str::operator== (char* otherChars) const
+{
+    return (Bool) !strcmp(operator char*(), otherChars);
+}
+
 Bool Str::eqNoCase (const char* other) const
 {
     return (strEqNoCase(operator char*(), other) ? TRUE : FALSE);
@@ -409,7 +414,7 @@ DStr Str::operator+ (int num) const
 // pack_() is overridden in DStr to put the pieces of the dynamic
 // string together
 //
-inline void Str::pack_() const
+void Str::pack_() const
 {
     if (!text_)
         const_cast<Str*>(this) -> empty();
@@ -439,26 +444,6 @@ void Str::speakTerse(DStr &ret)
 }
 
 
-// static function performing the same as isalnum() (so far).
-// should be extended to handle UTF-8 and XML's requirements
-// on valid name characters. FIXME: throw it out.
-//
-Bool Str::namechar(char c)
-{
-    return isalnum(c);
-}
-
-// another odd function (should be utf8isDigit())
-// FIXME: throw it out
-//
-Bool Str::isdigit_(char c)
-{
-    return (Bool) isdigit(c);
-}
-
-
-
-
 //
 //
 //                  DStr
@@ -578,7 +563,7 @@ void DStr::remove_()
     blocks.remove();   
 }
 
-inline void DStr::pack_() const
+void DStr::pack_() const
 {
     // if not needed leave asap
     if (blocks.isEmpty()) return;
@@ -593,12 +578,12 @@ inline void DStr::pack_() const
     // the length stays the same
 }
 
-inline int DStr::length() const
+int DStr::length() const
 {
     return byteLength_ + blocks.byteCount; 
 }
 
-inline DStr::operator char* () const
+DStr::operator char* () const
 {
     pack_();
     return Str::operator char*();
@@ -824,10 +809,10 @@ Str* NamespaceStack::getUri(const Str &_prefix) const
 Bool NamespaceStack::isHidden(const Str &_prefix) const
 {
     int ndx = findNum(_prefix);
-    return (ndx != -1) ? ((*this)[ndx] -> hide) : true;
+    return (ndx != -1) ? ((*this)[ndx] -> hide) : TRUE;
 }
 
-void NamespaceStack::appendConstruct(const Str &prefix, const Str& uri, Bool hide /* = false*/)
+void NamespaceStack::appendConstruct(const Str &prefix, const Str& uri, Bool hide /* = FALSE*/)
 {
     NamespaceStackObj* newItem = new NamespaceStackObj;
     newItem -> prefix  = prefix;
