@@ -73,7 +73,7 @@ public:
         // get last element
     inline T& last() const;
         // swap two elements with given indices
-    void swap(int,int);
+    virtual void swap(int,int);
     ~List(void);
 protected:
     void grow();
@@ -210,6 +210,7 @@ public:
         // NEEDS TO INCLUDE pack_()
     Bool operator== (const Str& other) const;
     Bool operator== (const char* otherChars) const;
+    Bool operator== (char* otherChars) const;
     
         // non-case-sensitive test for equality: to a Str, char*
         // NEEDS TO INCLUDE pack_()
@@ -234,8 +235,6 @@ public:
     // THE FOLLOWING 3 FUNCTIONS WOULD BE BEST REMOVED
         // output *this to another string with whitespace trimmed
     void speakTerse(DStr &);
-    static Bool namechar(char c);
-    static Bool isdigit_(char c);
 
     virtual ~Str();
 protected:
@@ -311,7 +310,7 @@ private:
 void escapeChars(DStr& result, const Str& what, 
                  const char* toEscape, const char** substitutes);
 
-extern const Str theEmptyString;
+/* extern const Str* theEmptyString; */
 
 /*****************************************************************
 NamespaceStack
@@ -332,7 +331,7 @@ public:
         // find the value of the element with the given key
     Str* getUri(const Str &) const;
     Bool isHidden(const Str &) const;
-    void appendConstruct(const Str &prefix, const Str& uri, Bool hide = false);
+    void appendConstruct(const Str &prefix, const Str& uri, Bool hide = FALSE);
 };
 
 /****************************************
@@ -371,7 +370,7 @@ public:
         // return the prefix
     const Str& getPrefix() const;
         // return true if a prefix is defined
-    const Bool QName::hasPrefix() const;
+    const Bool hasPrefix() const;
 
         // check for equality to another qname
     Bool operator==(const QName &) const;
@@ -483,17 +482,57 @@ class SList: public PList<T>
 public:
     SList(int logBlocksize_);
         // comparison of two elements
-    virtual int compare(T, T)
+    virtual int compare(int, int)
     { assert(!"abstract compare"); return 0; }
         // insertion of an element
     void insert(T);
-    void sort();
+    void sort(int from = 0, int to = - 1);  // to == -1 means "nItems-1"
 private:
     void quicksort(int bottom, int top);
     void qsPartition(int& i, int& j);
     void insertsort(int bottom, int top);
 };
 
+/***********
+
+  SortDef
+
+***********/
+
+class Expression;
+
+class SortDef
+{
+public:
+    Expression *sortExpr;
+    Str lang;
+    Bool asText, ascend, upper1st;
+    SortDef() 
+        : sortExpr(NULL), asText(TRUE), ascend(TRUE)
+        {};
+    SortDef(Expression *sortExpr_, Str lang_, 
+        Bool asText_, Bool ascend_, Bool upper1st_)
+        : sortExpr(sortExpr_), lang(lang_),
+          asText(asText_), ascend(ascend_), upper1st(upper1st_)
+        {};
+};
+
+/***********
+
+  SortDefList
+
+***********/
+
+class SortDefList : public PList<SortDef*>
+{
+    void appendConstruct(Expression *sortExpr_, Str lang_, 
+        Bool asText_, Bool ascend_, Bool upper1st_)
+    {
+        append(new SortDef(sortExpr_, lang_,
+            asText_, ascend_, upper1st_));
+    };
+};
+
 /*****************************************************************
 
 t e m p l a t e   m e t h o d s
@@ -649,24 +688,24 @@ PList<T>::PList(int logBlocksize_)
 template <class T>
 void PList<T>::freelast(Bool asArray)
 {
-    deleteScalarOrVector(last(), asArray);
-    deppend();
+    deleteScalarOrVector(List<T>::last(), asArray);
+    List<T>::deppend();
 }
 
 template <class T>
 void PList<T>::freeall(Bool asArray)
 {
-    for (int i = 0; i < nItems; i++)
-        deleteScalarOrVector(block[i], asArray);
-    deppendall();
+    for (int i = 0; i < List<T>::nItems; i++)
+        deleteScalarOrVector(List<T>::block[i], asArray);
+    List<T>::deppendall();
 }
 
 template <class T>
 void PList<T>::freerm(int n, Bool asArray)
 {
-    assert((n >= 0) && (n < nItems));
-    deleteScalarOrVector(block[n], asArray);
-    rm(n);
+    assert((n >= 0) && (n < List<T>::nItems));
+    deleteScalarOrVector(List<T>::block[n], asArray);
+    List<T>::rm(n);
 }
 
 /*================================================================
@@ -683,15 +722,15 @@ template <class T>
 void SList<T>::insert(T what)
 {
     append(what);
-    int whatndx = number() - 1;
+    int whatndx = List<T>::number() - 1;
     int i, j;
     for (i=0; i < whatndx; i++)
-        if (compare((*this)[whatndx],(*this)[i]) == -1) break;
+        if (compare(whatndx,i) == -1) break;
     if (i < whatndx)
     {
         for (j = whatndx; j > i; j--)
-            operator[](j) = operator[](j-1);
-        operator[](i) = what;
+            List<T>::operator[](j) = List<T>::operator[](j-1);
+        List<T>::operator[](i) = what;
     };
 };
 
@@ -702,10 +741,15 @@ void SList<T>::insertsort(int bottom, int top)
     T curr;
     for (ceiling = bottom + 1; ceiling <= top; ceiling++)
     {
-        curr = block[ceiling];
-        for (k = ceiling - 1; k >= bottom && compare(block[k], curr) > 0; k--)
-            block[k+1] = block[k];
-        block[k+1] = curr;
+        curr = List<T>::block[ceiling];
+        /* was:
+        for (k = ceiling - 1; k >= bottom && compare(k, ceiling) > 0; k--)
+            List<T>::block[k+1] = List<T>::block[k];
+        List<T>::block[k+1] = curr;
+         * but need to use swap() to drag extra pointers along
+         */
+        for (k = ceiling - 1; k >= bottom && compare(k, k + 1) > 0; k--)
+            swap(k, k+1); 
     }
 }
 
@@ -713,16 +757,25 @@ template <class T>
 void SList<T>::qsPartition(int& bottom, int& top)
 {
     int middle = (bottom + top) >> 1;
-    if (compare(block[bottom], block[top]) > 0) swap(bottom, top);
-    if (compare(block[middle], block[top]) > 0) swap(middle, top);
-    if (compare(block[bottom], block[middle]) > 0) swap(bottom, middle);
-    T pivot = block[middle];
+    if (compare(bottom, top) > 0) swap(bottom, top);
+    if (compare(middle, top) > 0) swap(middle, top);
+    if (compare(bottom, middle) > 0) swap(bottom, middle);
+    // T pivot = List<T>::block[middle];
     while (bottom <= top)
     {
-        while (compare(block[++bottom], pivot) <= 0);
-        while (compare(block[--top], pivot) >= 0);
-        if (bottom < top) swap(bottom, top);
-    };
+        // boundary conditions fixed by Tim
+        while ((++bottom <= top) && compare(bottom, middle) <= 0);
+        while ((--top >= bottom) && (compare(top, middle) >= 0));
+
+        if (bottom < top) 
+        {
+            if (middle == bottom)
+                middle = top;
+            else if (middle == top)
+                middle = bottom;
+            swap(bottom, top);
+        }
+    }
 }
 
 template <class T>
@@ -741,11 +794,14 @@ void SList<T>::quicksort(int bottom, int top)
 }
 
 template <class T>
-void SList<T>::sort()
+void SList<T>::sort(int from /* = 0 */, int to /* = - 1 */)
 {
-    if (nItems < 2)
+    /* cannot initialize 'to' to nItems-1 directly as g++ reports error */ 
+    if (to == -1)
+        to = List<T>::nItems - 1;
+    if (List<T>::nItems < 2)
         return;
-    quicksort(0, nItems - 1);
+    quicksort(from, to);
 }
 
 #endif //ifndef DataStrHIncl
