@@ -15,7 +15,7 @@
  * Portions created by Ginger Alliance are Copyright (C) 2000 Ginger
  * Alliance Ltd. All Rights Reserved.
  * 
- * Contributor(s):
+ * Contributor(s): Marc Lehmann <pcg@goof.com>
  * 
  * Alternatively, the contents of this file may be used under the
  * terms of the GNU General Public License Version 2 or later (the
@@ -36,6 +36,8 @@
 #include <math.h>
 #include "context.h"
 #include "tree.h"
+#include "utf8.h"
+
 //#include "proc.h"
 //#include <ctype.h>
 
@@ -61,7 +63,7 @@ ExToken tokenShortX[]=
 
 struct FuncInfoItem 
 {
-    char *name;
+    const char *name;
     ExFunctor func;
     ExType type;
 } 
@@ -352,7 +354,7 @@ eFlag Tokenizer::getToken_(ExToken &ret, char*& p, ExToken prevToken)
         case '&': assert(0);  //DBG: do not process entity references so far
             break;
         case '.':
-            if (Str::isdigit_(*(p+1)))
+            if (utf8IsDigit(utf8CharCode(p+1)))
             {
                 E( getNumber(p) );
                 ret = TOK_NUMBER;
@@ -364,14 +366,14 @@ eFlag Tokenizer::getToken_(ExToken &ret, char*& p, ExToken prevToken)
             break;
         default:
             {
-                if (Str::isdigit_(*p))
+                if (utf8IsDigit(utf8CharCode(p)))
                 {
                     E( getNumber(p) );
                     ret = TOK_NUMBER;
                 }
                 else
                 {
-                    if (Str::namechar(*p) || (*p == '_'))
+                    if (utf8IsLetter(utf8CharCode(p)) || (*p == '_') || (*p == ':'))
                     {
                         // the following call finds TOK_NAME, TOK_FNAME,
                         // TOK_AXISNAME,
@@ -405,13 +407,13 @@ eFlag Tokenizer::getToken(char*& p, TokenItem& item, ExToken prevToken)
 eFlag Tokenizer::getNumber(char*& p)
 {
     Bool wasDot = FALSE;
-    while ((*p) && (Str::isdigit_(*p) || (*p == '.')))
+    while ((*p) && (utf8IsDigit(utf8CharCode(p))) || (*p == '.'))
     {
         if (*p == '.')
             if (wasDot)
             Err(situation, ET_BAD_NUMBER)
             else wasDot = TRUE;
-        p++;
+        p += utf8SingleCharLength(p);
     };
     return OK;
 };
@@ -447,27 +449,33 @@ static Bool isNodeTest(char *p, int length)
     return (Bool)(q != NULL);
 }
 
-#define nameCharExtended(CH, PTR) ((CH = *PTR)!= 0) &&\
-        (Str::namechar(CH) || strchr(".-_:*",CH))
+#define nameCharExtended(CH, PTR) ((CH = utf8CharCode(PTR))!= 0) &&\
+        (utf8IsNameChar(CH) || strchr(".-_:*",CH))
 
 int nameLength(char* from)
 {
-    char *q = from, c;
-    while(nameCharExtended(c,q)) q++;
-    return (int)(q - from);
+    char *q = from;
+    int length = 0;
+    unsigned long c;
+    while(nameCharExtended(c,q)) {
+       q += utf8SingleCharLength(q);
+       length++;
+    }
+    return length;
 }
 
 eFlag Tokenizer::getName(ExToken &ret, char*& p, ExToken prevToken)
 {
     char *former = p;
-    char c;
+    unsigned long c;
     BOOL wasColon = FALSE;
     
-    if ((!Str::namechar(*p)) && (*p != '_'))
+    if ((!utf8IsLetter(utf8CharCode(p))) && (*p != '_'))
     {
         ret = TOK_NONE;
         return OK;
-    };
+    }
+
     while (nameCharExtended(c,p))
     {
         if (c == ':')
@@ -507,7 +515,7 @@ eFlag Tokenizer::getName(ExToken &ret, char*& p, ExToken prevToken)
                 return OK;
             }
         }
-        p++;
+        p += utf8SingleCharLength (p);
     }
 
     if (!wasColon && !isNamer(prevToken))
@@ -653,6 +661,7 @@ LocStep::LocStep(Processor* _proc,
 : preds(LIST_SIZE_1), proc(_proc), ntest(_proc)
 {
     set(_axis, _ntype);
+    positional = FALSE;
     badPreds = 0;
 }
 
@@ -786,11 +795,16 @@ eFlag LocStep::parse(Tokenizer& tokens, int& pos, Element *ownerV)
         {
             Expression *ex = new Expression(ownerV, tokens.proc);
             E( ex -> parse(tokens,i+1,right-1) );
-            // find out about the use of last()
-            if (ex -> optimizeUsesLast())
-                badPreds++;
-            // find sufficient position bounds
-            ex -> optimizePositionBounds();
+            // find out about the use of last() and position()
+            int exPositionalType = ex -> optimizePositional();
+            if (exPositionalType)
+            {
+                positional = TRUE;
+                if (exPositionalType == 2)
+                    badPreds++;
+                // find sufficient position bounds
+                ex -> optimizePositionBounds();
+            }
             preds.append(ex);
         };
         i = right + 1;
@@ -1270,7 +1284,8 @@ Bool Number::isInf()
 void Number::setNaN()
 {
     // divide by zero using a variable, to fool too clever compilers
-    int zero = 0;
+    // 'volatile' suggested by Dirk Siebnich
+    volatile int zero = 0;
     x = 0.0 / zero;
 }
 
@@ -1326,6 +1341,7 @@ Expression::Expression(Element *_ownerV,
     // is with it in a union
     patomnodeset = NULL;
     usesLast = FALSE;
+    positional = FALSE;
     optimizePositionFrom = optimizePositionTo = 0; // see header
 }
 
@@ -2098,13 +2114,20 @@ eFlag Expression::parse(Tokenizer& t, int from, int to)
                 E( e -> parse(t,from,leftmost - 1) );
                 int another = leftmost, 
                     lastone = leftmost;
-                t.getDelim(another);
+                // tom 24-10-00
+                // the following fails for "x and not(x and x)"
+                // t.getDelim(another);
+                // changing it to:
+                another = t.findTop(t.items[another]->tok, another+1);
                 while((another <= to) && (t.items[another]->tok != TOK_END))
                 {
                     args.append(e = new Expression(ownerV,proc));
                     E( e -> parse(t, lastone + 1, another - 1));
                     lastone = another;
-                    t.getDelim(another);
+
+                    // tom 14-11-00
+                    // t.getDelim(another);     failed too, for "x and x and (x and x)"
+                    another = t.findTop(t.items[another]->tok, another+1);
                 };
                 args.append(e = new Expression(ownerV,proc));
                 E( e -> parse(t,lastone + 1, to) );
@@ -2234,7 +2257,7 @@ eFlag Expression::eval(Expression &retxpr, Context *c)
         }; break;
     case EXF_FRAGMENT:
         {
-            newc = new Context;
+            newc = new Context(proc);
             newc -> set(&(pTree -> root));
             retxpr.setAtom(newc -> copy());
             delete newc;
@@ -2243,7 +2266,8 @@ eFlag Expression::eval(Expression &retxpr, Context *c)
         {
             int i, 
                 argsNumber = args.number();
-            ExprList atoms(argsNumber);
+            // was: ExprList atoms(argsNumber); (should be "log" if anything)
+            ExprList atoms(LIST_SIZE_1);
             Expression *ex;
             for (i = 0; i < argsNumber; i++)
             {
@@ -2281,6 +2305,20 @@ Bool hardCompare(ExFunctor op, T b1, T b2)
     return FALSE; //just to return something
 }
 
+ExFunctor _invertOp(ExFunctor op) 
+{
+    switch(op)
+    {
+    case EXFO_EQ: return EXFO_EQ; break;
+    case EXFO_NEQ: return EXFO_NEQ; break;
+    case EXFO_LT: return EXFO_GT; break;
+    case EXFO_GT: return EXFO_LT; break;
+    case EXFO_LE: return EXFO_GE; break;
+    case EXFO_GE: return EXFO_LE; break;
+    default: assert(!"_invertOp"); return EXF_NONE; // to return something
+    }
+}
+
 Bool Expression::compareCC(ExFunctor op, const Context &c1, const Context &c2)
 {
     DStr s1, s2;
@@ -2342,7 +2380,7 @@ Bool Expression::compareCN(ExFunctor op, const Context &c1, const Number& n2)
     {
         E( c -> current() -> value(s1, c) );
         n1 = s1;
-        if (::hardCompare(op, n1, (Number)n2))
+        if (hardCompare(op, n1, (Number)n2))
         {
             resulting = TRUE;
             break;
@@ -2387,16 +2425,16 @@ eFlag Expression::compare(Bool &result, Expression &other, ExFunctor op)
     else 
     {
         if (histype == EX_NODESET)
-            E( other.compare(result, *this, op) )
+            E( other.compare(result, *this, _invertOp(op)) )
         else
         {   // none of the two are nodesets
             if (type == EX_BOOLEAN || histype == EX_BOOLEAN)
                 result = hardCompare(op, tobool(), other.tobool());
             else
             {
-                if (type == EX_NUMBER || histype == EX_NUMBER)
-                    result = hardCompare(op, tonumber(), other.tonumber());
-                else
+                  if (type == EX_NUMBER || histype == EX_NUMBER)
+                      result = hardCompare(op, tonumber(), other.tonumber());
+                  else
                 {
                     if (type == EX_STRING || histype == EX_STRING)
                         result = hardCompare(op, tostring(), other.tostring());
@@ -2672,8 +2710,9 @@ eFlag Expression::callFunc(Expression &retxpr, ExprList &atoms, Context *c)
             checkArgsCount(2);
             checkIsString2(0,1);
             retxpr.setAtom((Bool) !firstOccurence(
-                atoms[0] -> tostringCharPtr(),
-                atoms[1] -> tostringCharPtr()));
+                // changing tostringCharPtr() to tostring():
+                atoms[0] -> tostring(),
+                atoms[1] -> tostring()));
         }; break;
 
     case EXFF_CONTAINS:
@@ -2681,20 +2720,23 @@ eFlag Expression::callFunc(Expression &retxpr, ExprList &atoms, Context *c)
             checkArgsCount(2);
             checkIsString2(0,1);
             retxpr.setAtom((Bool) (firstOccurence(
-                atoms[0] -> tostringCharPtr(),
-                atoms[1] -> tostringCharPtr()) != -1));
+                // changing tostringCharPtr() to tostring():
+                atoms[0] -> tostring(),
+                atoms[1] -> tostring()) != -1));
         }; break;
 
     case EXFF_SUBSTRING_BEFORE:
     case EXFF_SUBSTRING_AFTER:
         {
             Str s;
-            const Str& theSmaller = atoms[1] -> tostringRef();
-            char *thestring;
+            const Str& 
+                theBigger = atoms[0] -> tostring(),
+                theSmaller = atoms[1] -> tostring();
             checkArgsCount(2);
             checkIsString2(0,1);
             int where = firstOccurence(
-                thestring = atoms[0] -> tostringCharPtr(),
+                // changing tostringCharPtr() to tostring():
+                theBigger,
                 theSmaller);
             if (where == -1)
                 s.empty();
@@ -2705,10 +2747,10 @@ eFlag Expression::callFunc(Expression &retxpr, ExprList &atoms, Context *c)
                     if (where == 0)
                         s.empty();
                     else
-                        getBetween(s, thestring, 0, where-1);
+                        getBetween(s, theBigger, 0, where-1);
                 }
                 else
-                    getBetween(s, thestring, where + theSmaller.length(), -1);
+                    getBetween(s, theBigger, where + theSmaller.length(), -1);
             };
             retxpr.setAtom(s);
         }; break;
@@ -2795,10 +2837,14 @@ eFlag Expression::callFunc(Expression &retxpr, ExprList &atoms, Context *c)
             checkIsString(2);
 
             DStr resulting;
+            Str baseStr = atoms[0] -> tostring(),
+                srcStr = atoms[1] -> tostring(),
+                destStr = atoms[2] -> tostring();
             char *q,
-                *p = atoms[0] -> tostringCharPtr(),
-                *src = atoms[1] -> tostringCharPtr(),
-                *dest = atoms[2] -> tostringCharPtr();
+                // changing tostringCharPtr() to tostring():
+                *p = baseStr,
+                *src = srcStr,
+                *dest = destStr;
             int ndx,
                 destlen = strlen(dest);
             while(*p)
@@ -2912,7 +2958,7 @@ eFlag Expression::callFunc(Expression &retxpr, ExprList &atoms, Context *c)
             checkArgsCount(1);
             DStr location;
             Tree *newtree;
-            Context *newc = new Context;
+            Context *newc = new Context(proc);
 
             // Current node doesn't change
             newc->setCurrentNode (c->getCurrentNode());
@@ -2974,22 +3020,24 @@ eFlag Expression::callFunc(Expression &retxpr, ExprList &atoms, Context *c)
             q.setLogical(atoms[0] -> tostring(), &(NZ(ownerV) -> namespaces), FALSE);
             if (q.uri == proc -> stdPhrase(PHRASE_XSL_NAMESPACE))
             {
-                if (q.getLocal() == (char*) "version")
+                if (q.getLocal() == (const char*) "version")
                     retxpr.setAtom(Number(1.0));
-                else if (q.getLocal() == (char*) "vendor")
+                else if (q.getLocal() == (const char*) "vendor")
                     retxpr.setAtom(Str("Ginger Alliance"));
-                else if (q.getLocal() == (char*) "vendor-url")
+                else if (q.getLocal() == (const char*) "vendor-url")
                     retxpr.setAtom(Str("www.gingerall.com"));
                 else
-                    retxpr.setAtom(theEmptyString);
+                  retxpr.setAtom("");
+                /* retxpr.setAtom(*theEmptyString); */
             }
             else
-                retxpr.setAtom(theEmptyString);
+              retxpr.setAtom("");
+              /* retxpr.setAtom(*theEmptyString); */
         }; break;
 
     case EXFF_CURRENT:
         {
-            Context *newc = new Context;
+            Context *newc = new Context(proc);
             newc -> set (c -> getCurrentNode());
             retxpr.setAtom (newc);
         }; break;
@@ -3021,12 +3069,12 @@ eFlag Expression::createLPContext(Context *&c, int baseNdx)
 }
 */
 
-eFlag Expression::createLPContext(Context *&c, int baseNdx)
+eFlag Expression::createLPContext(Context *&c, int baseNdx, Vertex *givenGlobalCurrent /* = NULL */)
 {
     assert(functor == EXF_LOCPATH);
-    Context *theResult = new Context,
-        info;
-    info.setCurrentNode(c -> current());
+    Context *theResult = new Context(proc),
+        info(proc);
+    info.setCurrentNode(givenGlobalCurrent ? givenGlobalCurrent : c -> current());
     theResult -> setCurrentNode(c -> getCurrentNode());
     E( createLPContextLevel(0, args.number(), c -> current(), info, theResult) );
     theResult -> sort();
@@ -3062,7 +3110,8 @@ eFlag Expression::createLPContextLevel(
         totalReached(predsCount);   // serves as size for next (bad) pred
 
     // there will be as many dry (size-counting) runs as there are bad preds
-    Bool dryRun = TRUE;
+    Bool dryRun = TRUE,
+        quitThisRound = FALSE, quitThisVertex = FALSE;
     for (i = 0; i <= predsCount; i++)
     {
         if (i == predsCount)
@@ -3083,23 +3132,42 @@ eFlag Expression::createLPContextLevel(
             // locally current vertex 
             Vertex *locCurr = NULL;
 
+            quitThisRound = FALSE;
             do
             {
                 // shift the locally current vertex
                 E( args[stepLevel] -> step -> shift(locCurr, base) );
                 if (locCurr)
                 {
-                    if (lastBad < 0) ++reached[0];
+                    // was: if (lastBad < 0) ++reached[0];
+                    if ((lastBad < 0) || !dryRun) ++reached[0];
+                    quitThisVertex = FALSE;
                     for (j = 0; j < i; j++)
                     {
                         Bool satisfies;
                         info.deppendall();
                         info.setVirtual(locCurr, reached[j] - 1, totalReached[j]);
-                        E( args[stepLevel] -> step -> preds[j] -> 
-                          trueFor(&info, satisfies) );
-                        if (satisfies)
-                            ++reached[j + 1];
-                        else
+
+                        Expression *thisPred = 
+                            args[stepLevel] -> step -> preds[j];
+                        // find whether we're in position bounds for this pred
+                        switch(thisPred -> inBounds(reached[j] - 1))
+                        {
+                        case 0:
+                            {
+                                E( thisPred -> trueFor(&info, satisfies) );
+                                if (satisfies)
+                                    ++reached[j + 1];
+                                else
+                                    quitThisVertex = TRUE;
+                            }; break;
+                        case -1:
+                            break;
+                        case 1:
+                            quitThisRound = TRUE;
+                            break;
+                        };
+                        if (quitThisVertex || quitThisRound)
                             break;
                     };
                     if (j == i && !dryRun)     // passed all preds
@@ -3113,7 +3181,7 @@ eFlag Expression::createLPContextLevel(
 
                     }   // if ! dryRun
                 }       // if locCurr
-            } while (locCurr);          
+            } while (locCurr && !quitThisRound);
             // move all data collected to safe places
             for (init = lastBad + 1; init <= i; init++)
                 totalReached[init] = reached[init];
@@ -3124,16 +3192,16 @@ eFlag Expression::createLPContextLevel(
 }
 
 
-eFlag Expression::createLPContextSum(Context *&c)
+eFlag Expression::createLPContextSum(Context *&c, Vertex *globalCurrent /* = NULL */)
 {
     assert(functor == EXF_LOCPATH);
     Context 
-        *newc = new Context, 
+        *newc = new Context(proc), 
         *newc2, *returnedc;
     int cNumber = c -> getSize();
     for (int j = 0; j < cNumber; j++)
     {
-        E( createLPContext(returnedc = c, j) );
+        E( createLPContext(returnedc = c, j, globalCurrent) );
         newc2 = newc -> swallow(returnedc);
         delete newc;
         newc = newc2;
@@ -3204,10 +3272,15 @@ eFlag Expression::createContext(Context *& c, int baseNdx /* = -1 */)
         {
             assert(baseNdx != -1);  // meaningful only for a locpath
             E( args[0] -> createContext(newc = c, baseNdx) );
+
+            /// tom 21-10-00
+            /// wild guess:
+            newc -> setCurrentNode(c -> getCurrentNode());
+
             Context *filteredc;
             for (i = 1; i < argsNumber - (int) hasPath; i++)
             {
-                filteredc = new Context;
+                filteredc = new Context(proc);
                 newc -> reset();
                 Bool istrue;
                 int newcNumber = newc -> getSize();
@@ -3224,7 +3297,7 @@ eFlag Expression::createContext(Context *& c, int baseNdx /* = -1 */)
             };
             if (hasPath)
             {
-                E( args[argsNumber-1] -> createLPContextSum(filteredc = newc) );
+                E( args[argsNumber-1] -> createLPContextSum(filteredc = newc, newc -> getCurrentNode()) );
                 delete newc;
                 newc = filteredc;
             }
@@ -3238,7 +3311,7 @@ eFlag Expression::createContext(Context *& c, int baseNdx /* = -1 */)
 
             /////////
             // E( step -> createContextNoPreds(newc = c, baseNdx) );    - done as follows:
-            Context *newc = new Context;
+            Context *newc = new Context(proc);
             Vertex *curr = NULL;
             do
             {
@@ -3253,7 +3326,7 @@ eFlag Expression::createContext(Context *& c, int baseNdx /* = -1 */)
             int stepPredsNumber = step -> preds.number();
             for (i = 0; i < stepPredsNumber; i++)
             {
-                filteredc = new Context;
+                filteredc = new Context(proc);
                 newc -> reset();
                 Bool istrue;
                 int newcNumber = newc -> getSize();
@@ -3294,12 +3367,25 @@ eFlag Expression::matchesSingleStep(Vertex *v, Bool &result)
         RetOK(result, TRUE);
     if (!v -> parent)
         RetOK(result, FALSE);
-    Context *c = new Context, *newc;
-    c -> set(v -> parent);
-    E( createContext(newc = c, 0) );
-    result = (newc -> contains(v));
-    delete c;
-    delete newc;
+    if (!step -> positional)
+    {
+        Context *c = new Context(proc);
+        c -> set(v);
+        Bool stillOK = TRUE;
+        for (int i = 0; i < step -> preds.number() && stillOK; i++)
+            E(step -> preds[i] -> trueFor(c,stillOK));
+        delete c;
+        RetOK(result,stillOK);
+    }
+    else // positional case
+    {
+        Context *c = new Context(proc),*newc;
+        c -> set(v -> parent);
+        E( createContext(newc = c, 0) );
+        result = (newc -> contains(v));
+        delete c;
+        delete newc;
+    }
     return OK;
 }
 
@@ -3356,23 +3442,27 @@ eFlag Expression::matchesSinglePath(Vertex *v, int lastIndex, Bool& result)
 }
 
 /*
- *  optimizeUsesLast()
+ *  optimizePositional()
  *  called for a predicate of a locstep
- *  returns TRUE iff the predicate uses the last() function
- *  so one knows whether the size of the context has to be determined
+ *  returns 2 if the predicate uses the last() function
+ *    so the size of the context has to be determined
+ *  returns 1 if the predicate only uses position()
+ *          0 if neither
  */
 
-Bool Expression::optimizeUsesLast()
+int Expression::optimizePositional()
 {
-    Bool result = FALSE;
+    int result = 0;
     switch(functor)
     {
     case EXFF_LAST:
-        result = TRUE; break;
+        result = 2; break;
+    case EXFF_POSITION:
+        result = 1; break;
     case EXF_ATOM:
     case EXF_VAR:
     case EXF_LOCPATH:
-        /* result = FALSE; */ break;
+        /* result = 0; */ break;
     case EXF_FILTER:
     case EXF_STRINGSEQ:
     case EXF_FRAGMENT:
@@ -3381,16 +3471,20 @@ Bool Expression::optimizeUsesLast()
         break;
     default: // all the functions and operators, including EXF_OTHER_FUNC
         {
+            int sub = 0;
             for (int i = 0; i < args.number(); i++)
-                if (args[i] -> optimizeUsesLast())
+            {
+                if (!!(sub = args[i] -> optimizePositional()))
                 {
-                    result = TRUE;
-                    break;
+                    result = sub;
+                    if (result == 2) break;
                 }
+            }
         }
     }
-    usesLast = result;
-    return usesLast; 
+    usesLast = (result == 2);
+    positional = (result >= 1);
+    return result; 
 }
 
 /*
@@ -3435,3 +3529,12 @@ void Expression::optimizePositionBounds()
     optimizePositionFrom = from;
     optimizePositionTo = to;
 }
+
+int Expression::inBounds(int position) const
+{
+    if (optimizePositionTo && position > optimizePositionTo-1)
+        return 1;
+    if (optimizePositionFrom && position < optimizePositionFrom-1)
+        return -1;
+    return 0;
+}
