@@ -15,7 +15,7 @@
  * Portions created by Ginger Alliance are Copyright (C) 2000 Ginger
  * Alliance Ltd. All Rights Reserved.
  * 
- * Contributor(s):
+ * Contributor(s): Marc Lehmann <pcg@goof.com>
  * 
  * Alternatively, the contents of this file may be used under the
  * terms of the GNU General Public License Version 2 or later (the
@@ -38,6 +38,10 @@
 #include "uri.h"
 #include "proc.h"
 
+#ifdef HAVE_CONFIG_H
+#include <autocfg.h>
+#endif
+
 #define NONEMPTY_ELEMENT FALSE
 #define EMPTY_ELEMENT TRUE
 
@@ -84,6 +88,23 @@ const char* theEmptyHTML40Tags[] =
     "input", "isindex", "link", "meta", "param", NULL
 };
 
+const char* theURI_HTMLAtts[] = 
+{
+    "action",       // FORM
+    "archive",      // OBJECT
+    "background",   // BODY
+    "cite",         // BLOCKQUOTE, Q, DEL, INS
+    "classid",      // OBJECT
+    "codebase",     // OBJECT, APPLET
+    "data",         // OBJECT
+    "href",         // A, AREA, LINK, BASE
+    "longdesc",     // IMG, FRAME, IFRAME
+    "profile",      // HEAD
+    "src",          // SCRIPT, INPUT, FRAME, IFRAME, IMG
+    "usemap",       // IMG, INPUT, OBJECT
+    NULL
+};
+
 Bool isEmptyHTMLTag(const Str& name)
 {
     int ndx = lookupNoCase(name, theEmptyHTML40Tags);
@@ -91,6 +112,15 @@ Bool isEmptyHTMLTag(const Str& name)
     return !!theEmptyHTML40Tags[ndx];
 }
 
+//  this should be improved by checking the name of the parent element
+//  see the element names in theURI_HTMLAtts
+Bool isURI_HTMLAtt(const Str& name)
+{
+    int ndx = lookupNoCase(name, theURI_HTMLAtts);
+    // return TRUE iff found
+    return !!theURI_HTMLAtts[ndx];
+}
+
 //
 //  OutputDefinition
 //
@@ -203,12 +233,14 @@ Bool OutputDefinition::askQNameList(XSL_ATT itemId, const QName &what) const
 OutputMethod OutputDefinition::getMethod() const
 {
     const Str& method_ = getValueQName(XSLA_METHOD).getLocal();
-    if (method_ == (char*) "html")
+    if (method_ == (const char*) "html")
         return OUTPUT_HTML;
-    else if (method_ == (char*) "text")
+    else if (method_ == (const char*) "text")
         return OUTPUT_TEXT;
-    else if (method_ == (char*) "xml")
+    else if (method_ == (const char*) "xml")
         return OUTPUT_XML;
+    else if (method_ == (const char*) "xhtml")
+        return OUTPUT_XHTML;
     else return OUTPUT_UNKNOWN;
 }
 
@@ -261,6 +293,13 @@ eFlag OutputDefinition::setDefaults()
             E( setItemStrIfNot(XSLA_MEDIA_TYPE, "text/plain") );
             E( setItemStrIfNot(XSLA_OMIT_XML_DECL, strYes) );
         }; break;
+    case OUTPUT_XHTML:
+        {
+            E( setItemStrIfNot(XSLA_VERSION, "1.0") );
+            E( setItemStrIfNot(XSLA_INDENT, strYes) );
+            E( setItemStrIfNot(XSLA_MEDIA_TYPE, "text/html") );
+            E( setItemStrIfNot(XSLA_OMIT_XML_DECL, strYes) );
+        }; break;
     }
     return OK;
 }
@@ -294,13 +333,18 @@ eFlag PhysicalOutputLayerObj::setOptions(DataLine *targetDataLine_, OutputDefini
     method = outDef -> getMethod();
     if (method != OUTPUT_UNKNOWN)
         E( outDef -> setDefaults() );
-    enc = outDef -> getEncoding();
-    const Str& encStr = outDef -> getValueStr(XSLA_ENCODING);
+    //enc = outDef -> getEncoding();
+    Str encStr = proc -> getHardEncoding();
+    if (encStr.isEmpty())
+        encStr = outDef -> getValueStr(XSLA_ENCODING);
     if (!encStr.isEmpty() && !utf8CanRecodeTo(encStr))
     {
         Warn1(proc -> situation, W1_UNSUPP_OUT_ENCODING, encStr);
         E( outDef -> setItemStrForce(XSLA_ENCODING, "UTF-8") );
     }
+    else
+        E( outDef -> setItemStrForce(XSLA_ENCODING, encStr) );
+    enc = outDef -> getEncoding();
     return OK;
 }
 
@@ -353,10 +397,14 @@ eFlag PhysicalOutputLayerObj::outputElementStart(const Str& name,
             E( sendStr(attQName.getPrefix()) );
             sendLit(":");
         };
-        E( sendStr(atts[i] -> key.getLocal()) );
+        const Str& localAttName = atts[i] -> key.getLocal();
+        E( sendStr(localAttName) );
         sendLit("=\"");
-        E( sendStrEsc(atts[i] -> value, 
-            method == OUTPUT_HTML ? ESCAPING_HTML_ATTR : ESCAPING_ATTR) );
+        EscMode escapingMode = 
+            (method == OUTPUT_HTML ? ESCAPING_HTML_ATTR : ESCAPING_ATTR);
+        if (method == OUTPUT_HTML && isURI_HTMLAtt(localAttName))
+            escapingMode = ESCAPING_HTML_URI;
+        E( sendStrEsc(atts[i] -> value, escapingMode));
         sendLit("\"");
     };
 
@@ -366,9 +414,7 @@ eFlag PhysicalOutputLayerObj::outputElementStart(const Str& name,
         sendLit(">");
     else
     {
-        if (method != OUTPUT_HTML)
-            sendLit("/>");
-        else
+        if (method == OUTPUT_HTML || method == OUTPUT_XHTML)
         {
             if (!isEmptyHTMLTag(name))
             {
@@ -376,9 +422,13 @@ eFlag PhysicalOutputLayerObj::outputElementStart(const Str& name,
                 sendStr(name);
                 sendLit(">");
             }
-            else
+            else if (method == OUTPUT_HTML)
                 sendLit(">");
+            else
+                sendLit(" />");
         }
+        else
+            sendLit("/>");
     };
     return OK;
 }
@@ -408,8 +458,8 @@ eFlag PhysicalOutputLayerObj::outputComment(const Str& contents)
         E( sendOutUntil(p, len - (p - p_was), ESCAPING_NONE, "--") );
         if (*p)
         {
-            sendLit("- -");
-            p += 2;
+            sendLit("- ");
+            p++;
         }
     };
     if (trailingHyphen)
@@ -460,6 +510,7 @@ eFlag PhysicalOutputLayerObj::outputDTD(
     switch(method)
     {
     case OUTPUT_XML:
+    case OUTPUT_XHTML:
         {
             E( sendStr(name) );
             if (!systemId.isEmpty())
@@ -467,7 +518,7 @@ eFlag PhysicalOutputLayerObj::outputDTD(
                 if (!publicId.isEmpty())
                 {
                     sendLit(" PUBLIC \"");
-                    E( sendStrEsc(publicId, ESCAPING_URI) );
+                    E( sendStrEsc(publicId, ESCAPING_NONE) );
                     sendLit("\"");
                 }
                 else
@@ -483,7 +534,7 @@ eFlag PhysicalOutputLayerObj::outputDTD(
             if (!publicId.isEmpty())
             {
                 sendLit(" PUBLIC \"");
-                E( sendStrEsc(publicId, ESCAPING_URI) );
+                E( sendStrEsc(publicId, ESCAPING_NONE) );
                 sendLit("\"");
             }
             if (!systemId.isEmpty())
@@ -506,6 +557,7 @@ eFlag PhysicalOutputLayerObj::outputText(const Str& contents, Bool disableEsc,
     switch(method)
     {
     case OUTPUT_XML:
+    case OUTPUT_XHTML:
         {
             E( sendStrEsc(contents, disableEsc ? ESCAPING_NONE : ESCAPING_LT_AMP) );
         }; break;
@@ -548,7 +600,7 @@ int PhysicalOutputLayerObj::writeCharacterRef(char* dest, const char *src, EscMo
 eFlag PhysicalOutputLayerObj::sendOutUntil(const char *& data, int length,
     EscMode escapeMode, const char* stoppingText)
 {
-    char *p = strstr(data, stoppingText);
+    const char *p = strstr(data, stoppingText);
     int sending = (p ? p - data : length);
     E( sendOut(data, sending, escapeMode) );
     data += sending;
@@ -576,6 +628,7 @@ eFlag PhysicalOutputLayerObj::sendOut(const char* data, int length,
             {
                 switch(escapeMode)
                 {
+/*  will escape < in URIs as &lt; not the %xx way
                 case ESCAPING_URI:
                 case ESCAPING_HTML_URI:
                     {
@@ -584,6 +637,9 @@ eFlag PhysicalOutputLayerObj::sendOut(const char* data, int length,
                             ESCAPING_NONE) );
                         served = TRUE;
                     }; break;
+*/
+                case ESCAPING_URI:
+                case ESCAPING_HTML_URI:
                 case ESCAPING_ATTR:
                 case ESCAPING_LT_AMP:
                     {
@@ -596,19 +652,23 @@ eFlag PhysicalOutputLayerObj::sendOut(const char* data, int length,
             {
                 switch(escapeMode)
                 {
-                case ESCAPING_HTML_URI:
+/*  will escape & in URIs as &amp; not the %xx way
                 case ESCAPING_URI:
+                case ESCAPING_HTML_URI:
                     {
                         E( sendOut(smallBuf, 
                             writeCharacterRef(smallBuf, data, escapeMode),
                             ESCAPING_NONE) );
                         served = TRUE;
                     }; break;
+*/
                 case ESCAPING_HTML_ATTR:
                     {
                         if (data[1] == '{')
                             break;
                     }; // no break
+                case ESCAPING_URI:
+                case ESCAPING_HTML_URI:
                 case ESCAPING_ATTR:
                 case ESCAPING_LT_AMP:
                     {
@@ -621,6 +681,14 @@ eFlag PhysicalOutputLayerObj::sendOut(const char* data, int length,
             {
                 switch(escapeMode)
                 {
+                case ESCAPING_URI:
+                case ESCAPING_HTML_URI:
+                    {
+                        E( sendOut(smallBuf, 
+                            writeCharacterRef(smallBuf, data, escapeMode),
+                            ESCAPING_NONE) );
+                        served = TRUE;
+                    }; break;
                 case ESCAPING_HTML_ATTR:
                 case ESCAPING_ATTR:
                     {
@@ -647,6 +715,20 @@ eFlag PhysicalOutputLayerObj::sendOut(const char* data, int length,
                     }; break;
                 }
             }; break;
+        case ' ':
+            {
+                switch(escapeMode)
+                {
+                case ESCAPING_URI:
+                case ESCAPING_HTML_URI:
+                    {
+                        E( sendOut(smallBuf, 
+                            writeCharacterRef(smallBuf, data, escapeMode),
+                            ESCAPING_NONE) );
+                        served = TRUE;
+                    }; break;
+                }
+            }; break;
         };
         if (!served)
         {
@@ -703,6 +785,7 @@ eFlag PhysicalOutputLayerObj::setMethodByDefault(OutputMethod method_)
     switch( method = method_ )
     {
     case OUTPUT_XML:
+    case OUTPUT_XHTML:
         q.setLocal("xml"); break;
     case OUTPUT_HTML:
         q.setLocal("html"); break;
@@ -766,8 +849,8 @@ OutputterObj::OutputterObj(Processor *proc_)
     method = OUTPUT_UNKNOWN;
     noElementYet = TRUE;
     noHeadYet = TRUE;       // to recognize the first HEAD element
-    currNamespaces.appendConstruct("xsl",theXSLTNamespace,true);
-    currNamespaces.appendConstruct("xml",theXMLNamespace,true);
+    currNamespaces.appendConstruct("xsl",Str(theXSLTNamespace),TRUE);/* _PH_ */
+    currNamespaces.appendConstruct("xml",Str(theXMLNamespace),TRUE);/* _PH_ */
 }
 
 OutputterObj::~OutputterObj()
@@ -820,7 +903,7 @@ eFlag OutputterObj::reportXMLDeclIfMust()
 eFlag OutputterObj::reportDTDIfMust(const QName& docElementName)
 // to be called only after the output method is determined
 {
-    assert(method == OUTPUT_XML || method == OUTPUT_HTML);
+    assert(method != OUTPUT_TEXT);
     if (!physical)
         return OK;
     const Str& 
@@ -830,6 +913,7 @@ eFlag OutputterObj::reportDTDIfMust(const QName& docElementName)
     switch(method)
     {
     case OUTPUT_XML:
+    case OUTPUT_XHTML:
         writeDTD = !DTSystem.isEmpty();
         break;
     case OUTPUT_HTML:
@@ -913,7 +997,7 @@ eFlag OutputterObj::eventElementStart(const QName& name)
                 E( reportFront() );
             };
             // If this is an included stylesheet, output method cd be TEXT here
-            if (method == OUTPUT_XML || method == OUTPUT_HTML) {
+            if (method != OUTPUT_TEXT) {
                 E( reportDTDIfMust( name ) );
             }
         };
@@ -948,6 +1032,11 @@ eFlag OutputterObj::eventElementStart(const QName& name)
 eFlag OutputterObj::throwInMeta()
 {
     noHeadYet = FALSE;
+
+// define this symbol to disable adding the meta tag
+#ifdef SABLOT_DISABLE_ADDING_META
+    return OK;
+#endif
     Str metaName("meta");
     if (physical)
     {
@@ -1070,7 +1159,8 @@ eFlag OutputterObj::eventCommentEnd()
 {
     assert(state == STATE_IN_COMMENT);
     if (physical && method == OUTPUT_UNKNOWN)
-        E( front.appendConstruct(FM_COMMENT, currData, theEmptyString, FALSE) )
+        E( front.appendConstruct(FM_COMMENT, currData, 
+                                 ""/**theEmptyString*/, FALSE) )
     else
     {
         IF_PH2( outputComment, currData );
@@ -1147,7 +1237,8 @@ eFlag OutputterObj::eventData(const Str& data)
 {
     if (physical && method == OUTPUT_UNKNOWN && state == STATE_OUTSIDE)
     {
-        E( front.appendConstruct(FM_TEXT, data, theEmptyString, !outputEscaping) );
+        E( front.appendConstruct(FM_TEXT, data, 
+                                 "" /**theEmptyString*/, !outputEscaping) );
         if (!isAllWhite((const char*) data))
         {
             E( physical -> setMethodByDefault(method = OUTPUT_XML) );
@@ -1248,7 +1339,7 @@ eFlag OutputterObj::reportStartTag(Bool isEmpty)
     if (mySAXHandler)
     {
         const char**
-            attsTable = new const char*[(currAtts.number() << 1) + 1];
+            attsTable = new const char*[(currAtts.number() * 2) + 1];
         int i;
 
         // report only the new namespaces
@@ -1264,12 +1355,12 @@ eFlag OutputterObj::reportStartTag(Bool isEmpty)
         {
             Str *expatName = nameForSAX(currAtts[i] -> key);
             stringsForDeletion.append(expatName);
-            attsTable[i << 1] = *expatName;
-            attsTable[(i << 1) + 1] = currAtts[i] -> value;
+            attsTable[i * 2] = *expatName;
+            attsTable[(i * 2) + 1] = currAtts[i] -> value;
         };
-        attsTable[currAtts.number() << 1] = NULL;
-        
-        Str* theSAXname = nameForSAX(currElement); 
+        attsTable[currAtts.number() * 2] = NULL;
+
+        Str* theSAXname = nameForSAX(currElement);
         stringsForDeletion.append(theSAXname);
         IF_SAX3(startElement, *theSAXname, attsTable);
         delete[] attsTable;
