@@ -38,7 +38,6 @@
 #include "error.h"
 #include "proc.h"
 
-
 // for the windows-1250 encoding:
 #include "enc1250.h"
 // (will change to encodings.h)
@@ -204,8 +203,7 @@ eFlag TreeConstructer::feedDocumentToParser(void* constructer)
             XML_ParserFree(parser);
             return NOT_OK;
         };
-	quit = (Bool) (bytes == 0);
-        //quit = (Bool) (bytes < PARSE_BUFSIZE);
+        quit = (Bool) (bytes < PARSE_BUFSIZE);
         // res = XML_Parse(parser,buf,bytes,quit);
         res = XML_ParseBuffer(parser,bytes,quit);
         if (situation_.isError())
@@ -287,7 +285,9 @@ void TreeConstructer::tcStartElement(
         opCode = (XSL_OP) lookup((char*) q.getLocal(),xslOpNames);
         if (opCode == XSL_NONE)
         {
-            this_ -> theSituation -> error(ET_BAD_XSL, theEmptyString, theEmptyString);
+            this_ -> theSituation -> error(ET_BAD_XSL, 
+					   "" /**theEmptyString*/, 
+					   "" /**theEmptyString*/);
             return;
         };
         v = new(this_ -> theProcessor -> getArena()) XSLElement(q, t, opCode, this_ -> theProcessor);
@@ -314,7 +314,11 @@ void TreeConstructer::tcStartElement(
     };
     
     if (itsXSL)
+    {
         toX(v) -> checkAtts(); 
+        // also check if a toplevel element does not have a non-stylesheet parent
+        toX(v) -> checkToplevel(); 
+    }
     else
     {
         if (t -> XSLTree)
@@ -397,12 +401,13 @@ void TreeConstructer::tcStartNamespace(
     newv -> lineno = this_ -> getCurrentLineNumber();
     
     // warn on obsolete namespace
-    if (oldXSLTNamespace == uri)
+    if (uri && !strcmp(oldXSLTNamespace, uri)) /* _PH_ */
         Warn1(this_ -> theSituation, W1_OLD_NS_USED, (char*)uri)
     else
     {
-        if (prefix && !strcmp(prefix, "xsl") && !(theXSLTNamespace == uri))
-            Warn1(this_ -> theSituation, W1_XSL_NOT_XSL, (char*) uri);
+        if (prefix && !strcmp(prefix, "xsl") && 
+	      uri && strcmp(theXSLTNamespace, uri)) /* _PH_ */
+	  Warn1(this_ -> theSituation, W1_XSL_NOT_XSL, (char*) uri);
     }
 };
 
@@ -625,12 +630,22 @@ eFlag TreeConstructer::setQNameFromExpat(TreeConstructer* this_, QName& qname_,
         qname_.uri = UNDEF_PHRASE;
         qname_.setLocal(p);
         char *isColon;
+        qname_.prefix = UNDEF_PHRASE;
         if (!!(isColon = strchr(p,':')))
         {
             *isColon = 0;
-            Err1(this_ -> theProcessor -> situation, ET_BAD_PREFIX,p);
+
+            // fixing what appears a bug in expat - sometimes the xml:lang attr arrives unexpanded
+            // apparently only in XSL and when it's not at the top level
+            if (!strEqNoCase(p,"xml"))
+                Err1(this_ -> theProcessor -> situation, ET_BAD_PREFIX,p)
+            else
+            {
+                qname_.setLocal(isColon + 1);
+                qname_.uri = this_ -> theProcessor -> stdPhrase(PHRASE_XML_NAMESPACE);
+                E( this_ -> theProcessor -> dict().insert("xml",qname_.prefix) );                
+            }
         }
-        qname_.prefix = UNDEF_PHRASE;
     };
     return OK;
 }
