@@ -88,9 +88,9 @@ RuleSList::~RuleSList()
     freeall(FALSE);
 }
 
-int RuleSList::compare(RuleItem *first, RuleItem *second)
+int RuleSList::compare(int first, int second)
 {
-    return fcomp(first -> priority, second -> priority);
+    return fcomp((*this)[first] -> priority, (*this)[second] -> priority);
 };
 
 
@@ -288,7 +288,7 @@ eFlag Processor::open(char *sheetURI, char *inputURI, char *resultURI)
     double time_was = getMillisecs();
     situation->setCurrFile(styleSheet -> name);
 
-    Context *c = new Context;
+    Context *c = new Context(this);
     c -> set(&(input -> root));
 
     /*
@@ -638,7 +638,7 @@ const Str& Processor::findBaseURI(const Str& unmappedBase)
     uri2SchemePath(unmappedBase, scheme, rest);
     Str *mapping = baseURIMappings.find(scheme);
     if (mapping) return *mapping;
-    mapping = baseURIMappings.find(theEmptyString);
+    mapping = baseURIMappings.find(""/**theEmptyString*/);
     if (mapping) return *mapping;
     return unmappedBase;
 }
@@ -769,7 +769,7 @@ Element *findStylesheet(Daddy &d)
         if (isElement(w = d.contents[i]))
         {
             const QName& wName = toE(w) -> name;
-            if ((wName.getUri() == theXSLTNamespace) && 
+            if (!strcmp(wName.getUri(), theXSLTNamespace) && /* _PH_ */
               ((wName.getLocal() == xslOpNames[XSL_STYLESHEET]) ||
                (wName.getLocal() == xslOpNames[XSL_TRANSFORM])))
             return toE(w);
@@ -1015,6 +1015,17 @@ void *Processor::getHandlerUserData(HandlerType type, void *handler)
     }
 }
 
+void Processor::setHardEncoding(const Str& hardEncoding_)
+{
+    hardEncoding = hardEncoding_;
+}
+
+const Str& Processor::getHardEncoding() const
+{
+    return hardEncoding;
+};
+
+
 /*****************************************************************
 copyArg
   called if the result document's location uses the arg:
@@ -1081,7 +1092,7 @@ eFlag Processor::useGlobalParam(char *name, char *val)
 
 void Processor::setHardBaseURI(const char* hardBase)
 {
-    addBaseURIMapping(theEmptyString, (const Str) hardBase);
+  addBaseURIMapping(""/**theEmptyString*/, (const Str) hardBase);
 }
 
 void Processor::addBaseURIMapping(const Str& scheme, const Str& mapping)
@@ -1128,11 +1139,12 @@ Str Processor::getAliasedName(const QName& name, NamespaceStack& currNamespaces)
       } else {
         Str newPrefixStr = dictionary.getKey(newPrefix);
         s += newPrefixStr;
-        currNamespaces.appendConstruct(name.getPrefix(),name.getUri(),true);
+        currNamespaces.appendConstruct(name.getPrefix(),name.getUri(),TRUE);
         currNamespaces.appendConstruct(newPrefixStr,
                                        *currNamespaces.getUri(newPrefixStr));
       };
-      s += ":";
+      if (!s.isEmpty()) // could be empty e.g. for the default namespace
+        s += ":";
    };
    s += name.getLocal();
    return s;
