@@ -30,6 +30,10 @@
  * GPL.
  */
 
+#ifdef HAVE_CONFIG_H
+#include <autocfg.h>
+#endif
+
 #include "verts.h"
 #include "tree.h"
 #include "expr.h"
@@ -894,7 +898,7 @@ eFlag NmSpace::execute(Context *c)
 */
     E( proc -> outputter() -> 
         eventNamespace(
-	        proc -> dict().getKey(prefix), //was: proc -> aliasXL(prefix)),
+                proc -> dict().getKey(prefix), //was: proc -> aliasXL(prefix)),
             proc -> dict().getKey(uri)) );
     return OK;
 }
@@ -1185,6 +1189,13 @@ eFlag XSLElement::execute(Context *c)
             E( e -> createContext( newc ) ); 
             if (!newc -> isVoid())
             {
+                // *** SORT HERE ***
+                SortDefList sortDefs;
+                makeSortDefs(sortDefs, c);
+                if (sortDefs.number())
+                    E( newc -> sort(this, &sortDefs) );
+                sortDefs.freeall(FALSE);
+
                 // process with-params
                 E( contents.execute(c) );
                 E( proc -> execute((Vertex*) NULL, newc) );
@@ -1366,6 +1377,13 @@ eFlag XSLElement::execute(Context *c)
             E( NZ( a -> expr ) -> createContext(newc));
             if (!newc -> isVoid() && contents.number())
             {
+                // *** SORT HERE ***
+                SortDefList sortDefs;
+                makeSortDefs(sortDefs, c);
+                if (sortDefs.number())
+                    E( newc -> sort(this, &sortDefs) );
+                sortDefs.freeall(FALSE);
+
                 E( proc -> execute(contents, newc) );
                 // E( proc -> execute(contents[0],newc) );
                 // proc disposes of the new context
@@ -1398,7 +1416,7 @@ eFlag XSLElement::execute(Context *c)
            Expression *expr = new Expression(this, proc);
            if (contents.isEmpty())
            {
-              expr -> setAtom(theEmptyString);
+             expr -> setAtom(""/**theEmptyString*/);
            } else {
               proc -> vars -> startNested();
               TreeConstructer *newTC;
@@ -1449,7 +1467,7 @@ eFlag XSLElement::execute(Context *c)
                 freeit = TRUE;
                 expr = new Expression(this, proc);
                 if (contents.isEmpty())
-                    expr -> setAtom(theEmptyString);
+                  expr -> setAtom(""/**theEmptyString*/);
                 else // result tree fragment
                 {
                     // starting a nesting will make the current prebindings
@@ -1499,6 +1517,9 @@ eFlag XSLElement::execute(Context *c)
     case XSL_NAMESPACE_ALIAS:
         // these were processed during parse
         break;
+    case XSL_SORT:
+        // processed by parent
+        break;
     // unsupported instructions that are not considered harmful
     case XSL_STRIP_SPACE:
     case XSL_PRESERVE_SPACE:
@@ -1562,8 +1583,10 @@ eFlag XSLElement::checkAtts()
     for (int i = 0; i < atts.number(); i++)
     {
         a = cast(Attribute*, atts[i]);
-        if ((attNdx = findAttNdx(instrData,a)) == -1)
-            Err1(proc -> situation, ET_BAD_ATTR,a -> name.getname());
+        if (((attNdx = findAttNdx(instrData,a)) == -1) && 
+            a -> name.uri == UNDEF_PHRASE)
+            // FIXME: issue a warning when ignoring an att??
+            Err1(proc -> situation, ET_BAD_ATTR,a -> name.getname())
         if (instrData.att[attNdx].required)
             reqCount++;
         if (instrData.att[attNdx].exprType != EX_NONE)
@@ -1632,6 +1655,16 @@ void XSLElement::checkExtraChildren(int& k)
     }
 }
 
+eFlag XSLElement::checkToplevel()
+{
+    if (!(instrTable[op].flags & ELEM_INSTR) && 
+        !(instrTable[op].flags & ELEM_EXTRA) &&
+        (!isXSL(parent) || 
+         !(instrTable[toX(parent) -> op].flags & ELEM_CONT_TOPLEVEL)))
+        Err1(proc -> situation, E1_ELEM_TOPLEVEL, xslOpNames[op]);
+    return OK;
+}
+
 eFlag XSLElement::checkChildren()
 {
     InstrTableItem &iData = instrTable[op];
@@ -1669,7 +1702,107 @@ eFlag XSLElement::checkChildren()
                 Err1(proc -> situation, E_BAD_ELEM_CONTENT, xslOpNames[op]);
         }
     };
+
     proc -> situation->popCurrent();
     return OK;
 }
 
+eFlag XSLElement::make1SortDef(SortDef *&def, Context *c)
+{
+    DStr temp;
+
+    assert(op == XSL_SORT);
+    def = new SortDef;
+    
+    Attribute *a = atts.find(XSLA_SELECT);
+    if (a)
+        def -> sortExpr = a -> expr;
+
+    a = atts.find(XSLA_LANG);
+    if (a)
+    {
+        E( a -> value(temp, c));
+        def -> lang = temp;
+    }
+    else
+    {
+        // lang is system-dependent!!! How about setting ""?
+        def -> lang = "en";
+    };
+
+    a = atts.find(XSLA_DATA_TYPE);
+    if (a)
+    {
+        E( a -> value(temp, c) );
+        if (temp == (const char*) "number")
+            def -> asText = FALSE;
+        else
+            if (!(temp == (const char*) "text"))
+                Warn1(proc -> situation, W1_SORT_DATA_TYPE, temp);
+    }
+
+    a = atts.find(XSLA_ORDER);
+    if (a)
+    {
+        E( a -> value(temp, c) );
+        if (temp == (const char*) "descending")
+            def -> ascend = FALSE;
+        else if (!(temp == (const char*) "ascending"))
+            Warn1(proc -> situation, W1_SORT_ORDER, temp);
+    }
+    
+    a = atts.find(XSLA_CASE_ORDER);
+    if (a)
+    {
+        E( a -> value(temp, c) );
+        if (temp == (const char*) "lower-first")
+            def -> upper1st = FALSE;
+        else 
+        {
+            if (!(temp == (const char*) "upper-first"))
+                Warn1(proc -> situation, W1_SORT_CASE_ORDER, temp);
+            // get upper1st as system default!!
+            def -> upper1st = TRUE;
+        }
+    }
+    return OK;
+}
+
+eFlag XSLElement::makeSortDefs(SortDefList &sortDefs, Context *c)
+{
+    assert(op == XSL_APPLY_TEMPLATES || op == XSL_FOR_EACH);
+
+    // tom: Jan 7, 01
+    // adding a test for wcsxfrm()
+    // ignore sort and warn if unavailable
+/*
+#   if !defined(HAVE_WCSXFRM) && !defined(_MSC_VER)
+    Warn(proc -> situation, W_NO_WCSXFRM);
+    // return with empty sortDefs
+    return OK;
+#   endif
+*/
+
+    Vertex *child;
+    for (int i = 0; i < contents.number(); i++)
+    {
+        child = contents[i];
+        if (!isXSLElement(child)) break;
+        if (toX(child) -> op == XSL_SORT)
+        {
+            // *** make the single sort def
+            SortDef *newDef;
+            E( toX(child) -> make1SortDef(newDef, c));
+            sortDefs.append(newDef);
+#           if !defined(HAVE_WCSXFRM) && !defined(_MSC_VER)
+            // warn that string sorting may be incorrect - 
+            // sorting as ascii
+            if (newDef -> asText)
+                Warn(proc -> situation, W_NO_WCSXFRM);
+#           endif
+
+        }
+        else if (toX(child) -> op != XSL_WITH_PARAM) break;
+    };
+    return OK;
+}
